from typing import TYPE_CHECKING
from rid_lib.ext import Cache

from ..network.resolver import NetworkResolver
from ..config.core import NodeConfig
from ..config.loader import ConfigLoader
from ..network.graph import NetworkGraph
from ..network.event_queue import EventQueue
from ..network.request_handler import RequestHandler
from ..identity import NodeIdentity
from .kobj_queue import KobjQueue

if TYPE_CHECKING:
    from ..effector import Effector


class HandlerContext:
    """Context object provides knowledge handlers access to other components."""
    
    identity: NodeIdentity
    config: NodeConfig
    config_loader: ConfigLoader
    cache: Cache
    event_queue: EventQueue
    kobj_queue: KobjQueue
    graph: NetworkGraph
    request_handler: RequestHandler
    resolver: NetworkResolver
    effector: "Effector"
    
    def __init__(
        self,
        identity: NodeIdentity,
        config: NodeConfig,
        config_loader: ConfigLoader,
        cache: Cache,
        event_queue: EventQueue,
        kobj_queue: KobjQueue,
        graph: NetworkGraph,
        request_handler: RequestHandler,
        resolver: NetworkResolver
    ):
        self.identity = identity
        self.config = config
        self.config_loader = config_loader
        self.cache = cache
        self.event_queue = event_queue
        self.kobj_queue = kobj_queue
        self.graph = graph
        self.request_handler = request_handler
        self.resolver = resolver
    
    def set_effector(self, effector: "Effector"):
        """Post initialization injection of effector component."""
        self.effector = effector


import structlog
import uvicorn
from contextlib import asynccontextmanager
from fastapi import FastAPI, APIRouter
from fastapi.responses import JSONResponse

from .base import EntryPoint
from ..network.response_handler import ResponseHandler
from ..protocol.model_map import API_MODEL_MAP
from ..protocol.api_models import ErrorResponse
from ..protocol.errors import ProtocolError
from ..lifecycle import NodeLifecycle
from ..config.full_node import FullNodeConfig

log = structlog.stdlib.get_logger()


class NodeServer(EntryPoint):
    """Entry point for full nodes, manages FastAPI server."""
    config: FullNodeConfig
    lifecycle: NodeLifecycle
    response_handler: ResponseHandler
    app: FastAPI
    router: APIRouter
    
    def __init__(
        self,
        config: FullNodeConfig,
        lifecycle: NodeLifecycle,
        response_handler: ResponseHandler,
    ):
        self.config = config
        self.lifecycle = lifecycle
        self.response_handler = response_handler
        
    def build_endpoints(self, router: APIRouter):
        """Builds endpoints for API router."""
        for path, models in API_MODEL_MAP.items():
            def create_endpoint(path: str):
                async def endpoint(req):
                    return self.response_handler.handle_response(path, req)
                
                # programmatically setting type hint annotations for FastAPI's model validation 
                endpoint.__annotations__ = {
                    "req": models.request_envelope,
                    "return": models.response_envelope
                }
                
                return endpoint
            
            router.add_api_route(
                path=path,
                endpoint=create_endpoint(path),
                methods=["POST"],
                response_model_exclude_none=True
            )
    
    def build_app(self):
        """Builds FastAPI app."""
        self.app = FastAPI(
            title="KOI-net Protocol API",
            version="1.1.0"
        )
        
        self.app.add_exception_handler(ProtocolError, self.protocol_error_handler)
        self.router = APIRouter(prefix="/koi-net")
        self.build_endpoints(self.router)
        self.app.include_router(self.router)
        
    def protocol_error_handler(self, request, exc: ProtocolError):
        """Catches `ProtocolError` and returns an `ErrorResponse` payload."""
        log.error(exc)
        resp = ErrorResponse(error=exc.error_type)
        log.info(f"Returning error response: {resp}")
        return JSONResponse(
            status_code=400,
            content=resp.model_dump(mode="json")
        )
    
    def run(self):
        """Starts FastAPI server and event handler."""
        
        with self.lifecycle.run():
            self.build_app()
            
            uvicorn.run(
                app=self.app,
                host=self.config.server.host,
                port=self.config.server.port,
                log_config=None,
                access_log=False
            )



import time
import structlog

from .base import EntryPoint
from ..processor.kobj_queue import KobjQueue
from ..lifecycle import NodeLifecycle
from ..network.resolver import NetworkResolver
from ..config.partial_node import PartialNodeConfig

log = structlog.stdlib.get_logger()


class NodePoller(EntryPoint):
    """Entry point for partial nodes, manages polling event loop."""
    kobj_queue: KobjQueue
    lifecycle: NodeLifecycle
    resolver: NetworkResolver
    config: PartialNodeConfig
    
    def __init__(
        self,
        config: PartialNodeConfig,
        lifecycle: NodeLifecycle,
        kobj_queue: KobjQueue,
        resolver: NetworkResolver,
    ):
        self.kobj_queue = kobj_queue
        self.lifecycle = lifecycle
        self.resolver = resolver
        self.config = config

    def poll(self):
        """Polls neighbor nodes and processes returned events."""
        for node_rid, events in self.resolver.poll_neighbors().items():
            for event in events:
                self.kobj_queue.push(event=event, source=node_rid)

    def run(self):
        """Runs polling event loop."""
        with self.lifecycle.run():
            while True:
                start_time = time.time()
                self.poll()
                elapsed = time.time() - start_time
                sleep_time = self.config.poller.polling_interval - elapsed
                if sleep_time > 0:
                    time.sleep(sleep_time)


from rid_lib.types import KoiNetNode, KoiNetEdge
import structlog
from koi_net.config.full_node import (
    FullNodeConfig, 
    ServerConfig, 
    KoiNetConfig, 
    NodeProfile, 
    NodeProvides
)
from koi_net.core import FullNode
from koi_net.processor.context import HandlerContext
from koi_net.processor.handler import HandlerType, KnowledgeHandler
from koi_net.processor.knowledge_object import KnowledgeObject
from koi_net.protocol.event import Event, EventType
from koi_net.protocol.edge import EdgeType, generate_edge_bundle

log = structlog.stdlib.get_logger()


class CoordinatorConfig(FullNodeConfig):
    server: ServerConfig = ServerConfig(port=8080)
    koi_net: KoiNetConfig = KoiNetConfig(
        node_name="coordinator",
        node_profile=NodeProfile(
            provides=NodeProvides(
                event=[KoiNetNode, KoiNetEdge],
                state=[KoiNetNode, KoiNetEdge]
            )
        ),
        rid_types_of_interest=[KoiNetNode, KoiNetEdge]
    )

@KnowledgeHandler.create(
    HandlerType.Network, 
    rid_types=[KoiNetNode])
def handshake_handler(ctx: HandlerContext, kobj: KnowledgeObject):
    log.info("Handling node handshake")

    # only respond if node declares itself as NEW
    if kobj.event_type != EventType.NEW:
        return
        
    log.info("Sharing this node's bundle with peer")
    identity_bundle = ctx.cache.read(ctx.identity.rid)
    ctx.event_queue.push(
        event=Event.from_bundle(EventType.NEW, identity_bundle),
        target=kobj.rid
    )
    
    log.info("Proposing new edge")    
    # defer handling of proposed edge
    
    edge_bundle = generate_edge_bundle(
        source=kobj.rid,
        target=ctx.identity.rid,
        edge_type=EdgeType.WEBHOOK,
        rid_types=[KoiNetNode, KoiNetEdge]
    )
        
    ctx.kobj_queue.push(rid=edge_bundle.rid, event_type=EventType.FORGET)
    ctx.kobj_queue.push(bundle=edge_bundle)

class CoordinatorNode(FullNode):
    config_schema = CoordinatorConfig
    knowledge_handlers = FullNode.knowledge_handlers + [handshake_handler]

if __name__ == "__main__":
    node = CoordinatorNode()
    node.entrypoint.run()


from typing import Callable

from rid_lib import RID
from rid_lib.ext import Bundle, Cache

cache = Cache()

def build_dereferencer(
    *funcs: list[Callable[[RID], Bundle | None]]
) -> Callable[[RID], Bundle | None]:
    def any_of(rid: RID):
        return any(
            f(rid) for f in funcs
        )
    return any_of

deref = build_dereferencer(cache.read)
deref(RID.from_string("string:hello_world"))


import structlog
from rid_lib.ext import Cache
from rid_lib.types import KoiNetNode
from koi_net.identity import NodeIdentity
from koi_net.network.event_queue import EventQueue
from .protocol.event import Event, EventType

log = structlog.stdlib.get_logger()


class Handshaker:
    """Handles handshaking with other nodes."""
    def __init__(
        self, 
        cache: Cache, 
        identity: NodeIdentity, 
        event_queue: EventQueue
    ):
        self.cache = cache
        self.identity = identity
        self.event_queue = event_queue
        
    def handshake_with(self, target: KoiNetNode):
        """Initiates a handshake with target node.
        
        Pushes successive `FORGET` and `NEW` events to target node to
        reset the target's cache in case it already knew this node. 
        """
        
        log.debug(f"Initiating handshake with {target}")
        self.event_queue.push(
            Event.from_rid(
                event_type=EventType.FORGET, 
                rid=self.identity.rid),
            target=target
        )
        self.event_queue.push(
            event=Event.from_bundle(
                event_type=EventType.NEW, 
                bundle=self.cache.read(self.identity.rid)),
            target=target
        )


from .cache import Cache
from .log_system import LogSystem
from .assembler import NodeAssembler
from .config.core import NodeConfig
from .config.proxy import ConfigProxy
from .config.loader import ConfigLoader
from .processor.context import HandlerContext
from .effector import DerefHandler, Effector
from .handshaker import Handshaker
from .sync_manager import SyncManager
from .identity import NodeIdentity
from .workers import KnowledgeProcessingWorker, EventProcessingWorker
from .lifecycle import NodeLifecycle
from .network.error_handler import ErrorHandler
from .network.event_queue import EventQueue
from .network.graph import NetworkGraph
from .network.request_handler import RequestHandler
from .network.resolver import NetworkResolver
from .network.response_handler import ResponseHandler
from .network.event_buffer import EventBuffer
from .processor.pipeline import KnowledgePipeline
from .processor.kobj_queue import KobjQueue
from .processor.handler import KnowledgeHandler
from .secure_manager import SecureManager
from .entrypoints import NodeServer, NodePoller
from .processor.knowledge_handlers import (
    basic_manifest_handler, 
    basic_network_output_filter, 
    basic_rid_handler, 
    node_contact_handler, 
    edge_negotiation_handler, 
    forget_edge_on_node_deletion, 
    secure_profile_handler
)

class BaseNode(NodeAssembler):
    log_system: LogSystem = LogSystem
    kobj_queue: KobjQueue = KobjQueue
    event_queue: EventQueue = EventQueue
    poll_event_buf: EventBuffer = EventBuffer
    broadcast_event_buf: EventBuffer = EventBuffer
    config_schema = NodeConfig
    config: NodeConfig = ConfigProxy
    config_loader: ConfigLoader = ConfigLoader
    knowledge_handlers: list[KnowledgeHandler] = [
        basic_rid_handler,
        basic_manifest_handler,
        secure_profile_handler,
        edge_negotiation_handler,
        node_contact_handler,
        basic_network_output_filter,
        forget_edge_on_node_deletion
    ]
    deref_handlers: list[DerefHandler] = []
    cache: Cache = Cache
    identity: NodeIdentity = NodeIdentity
    graph: NetworkGraph = NetworkGraph
    secure_manager: SecureManager = SecureManager
    handshaker: Handshaker = Handshaker
    error_handler: ErrorHandler = ErrorHandler
    request_handler: RequestHandler = RequestHandler
    sync_manager: SyncManager = SyncManager
    response_handler: ResponseHandler = ResponseHandler
    resolver: NetworkResolver = NetworkResolver
    handler_context: HandlerContext = HandlerContext
    effector: Effector = Effector
    pipeline: KnowledgePipeline = KnowledgePipeline
    kobj_worker: KnowledgeProcessingWorker = KnowledgeProcessingWorker
    event_worker: EventProcessingWorker = EventProcessingWorker
    lifecycle: NodeLifecycle = NodeLifecycle

class FullNode(BaseNode):
    entrypoint: NodeServer = NodeServer

class PartialNode(BaseNode):
    entrypoint: NodePoller = NodePoller


import structlog
from rid_lib.types.koi_net_node import KoiNetNode
from .config.core import NodeConfig
from .protocol.node import NodeProfile

log = structlog.stdlib.get_logger()

    
class NodeIdentity:
    """Represents a node's identity (RID, profile)."""
    
    config: NodeConfig
    
    def __init__(self, config: NodeConfig):
        self.config = config
        
    @property
    def rid(self) -> KoiNetNode:
        return self.config.koi_net.node_rid
    
    @property
    def profile(self) -> NodeProfile:
        return self.config.koi_net.node_profile



from koi_net.config.core import NodeConfig


class ConfigProxy:
    """Proxy for config access.
    
    Allows initialization of this component, and updating state without
    destroying the original reference. Handled as if it were a config
    model by other classes, loaded and saved by the `ConfigLoader`.
    """
    _config: NodeConfig
    
    def __init__(self):
        self._config = None
    
    def __getattr__(self, name):
        if not self._config:
            raise Exception("Proxy called before config loaded")
            
        return getattr(self._config, name)


from rid_lib.ext import Cache

from .network.graph import NetworkGraph
from .network.request_handler import RequestHandler
from .processor.kobj_queue import KobjQueue
from .protocol.api_models import ErrorResponse
from .protocol.node import NodeProfile, NodeType


class SyncManager:
    """Handles state synchronization actions with other nodes."""
    graph: NetworkGraph
    cache: Cache
    request_handler: RequestHandler
    kobj_queue: KobjQueue
    
    def __init__(
        self,
        graph: NetworkGraph,
        cache: Cache,
        request_handler: RequestHandler,
        kobj_queue: KobjQueue
    ):
        self.graph = graph
        self.cache = cache
        self.request_handler = request_handler
        self.kobj_queue = kobj_queue
    
    def catch_up_with(self, nodes, rid_types):
        """Catches up with the state of RID types within other nodes."""
    
        for node in nodes:
            node_bundle = self.cache.read(node)
            node_profile = node_bundle.validate_contents(NodeProfile)
            
            # can't catch up with partial nodes
            if node_profile.node_type != NodeType.FULL:
                continue
            
            payload = self.request_handler.fetch_manifests(
                node, rid_types=rid_types)
            
            if type(payload) is ErrorResponse:
                continue
            
            for manifest in payload.manifests:
                self.kobj_queue.push(
                    manifest=manifest,
                    source=node
                )


import structlog
from contextlib import contextmanager, asynccontextmanager

from rid_lib.ext import Bundle
from rid_lib.types import KoiNetNode

from koi_net.config.loader import ConfigLoader
from koi_net.secure_manager import SecureManager

from .sync_manager import SyncManager
from .handshaker import Handshaker
from .workers.kobj_worker import KnowledgeProcessingWorker
from .network.event_queue import EventQueue
from .workers import EventProcessingWorker
from .workers.base import STOP_WORKER
from .config.core import NodeConfig
from .processor.kobj_queue import KobjQueue
from .network.graph import NetworkGraph
from .identity import NodeIdentity

log = structlog.stdlib.get_logger()


class NodeLifecycle:
    """Manages node startup and shutdown processes."""
    
    config: NodeConfig
    config_loader: ConfigLoader
    identity: NodeIdentity
    graph: NetworkGraph
    kobj_queue: KobjQueue
    kobj_worker: KnowledgeProcessingWorker
    event_queue: EventQueue
    event_worker: EventProcessingWorker
    handshaker: Handshaker
    sync_manager: SyncManager
    secure_manager: SecureManager
    
    def __init__(
        self,
        config: NodeConfig,
        config_loader: ConfigLoader,
        identity: NodeIdentity,
        graph: NetworkGraph,
        kobj_queue: KobjQueue,
        kobj_worker: KnowledgeProcessingWorker,
        event_queue: EventQueue,
        event_worker: EventProcessingWorker,
        handshaker: Handshaker,
        sync_manager: SyncManager,
        secure_manager: SecureManager
    ):
        self.config = config
        self.config_loader = config_loader
        self.identity = identity
        self.graph = graph
        self.kobj_queue = kobj_queue
        self.kobj_worker = kobj_worker
        self.event_queue = event_queue
        self.event_worker = event_worker
        self.handshaker = handshaker
        self.sync_manager = sync_manager
        self.secure_manager = secure_manager
        
    @contextmanager
    def run(self):
        """Synchronous context manager for node startup and shutdown."""
        try:
            log.info("Starting node lifecycle...")
            self.start()
            yield
        except KeyboardInterrupt:
            log.info("Keyboard interrupt!")
        finally:
            log.info("Stopping node lifecycle...")
            self.stop()

    @asynccontextmanager
    async def async_run(self):
        """Asynchronous context manager for node startup and shutdown."""
        try:
            log.info("Starting async node lifecycle...")
            self.start()
            yield
        except KeyboardInterrupt:
            log.info("Keyboard interrupt!")
        finally:
            log.info("Stopping async node lifecycle...")
            self.stop()
    
    def start(self):
        """Starts a node.
        
        Starts the processor thread (if enabled). Generates network 
        graph from nodes and edges in cache. Processes any state changes 
        of node bundle. Initiates handshake with first contact if node 
        doesn't have any neighbors. Catches up with coordinator state.
        """
        
        # attempt to load config from yaml, and write back changes (if any)
        self.config_loader.load_from_yaml()
        self.config_loader.save_to_yaml()
        
        self.secure_manager.load_priv_key()
        
        self.kobj_worker.thread.start()
        self.event_worker.thread.start()
        self.graph.generate()
        
        # refresh to reflect changes (if any) in config.yaml node profile
        self.kobj_queue.push(bundle=Bundle.generate(
            rid=self.identity.rid,
            contents=self.identity.profile.model_dump()
        ))
        
        self.kobj_queue.q.join()
        
        node_providers = self.graph.get_neighbors(
            direction="in",
            allowed_type=[KoiNetNode]
        )
        
        if node_providers:
            log.debug(f"Catching up with `orn:koi-net.node` providers: {node_providers}")
            self.sync_manager.catch_up_with(node_providers, [KoiNetNode])
        
        elif self.config.koi_net.first_contact.rid:
            log.debug(f"No edges with `orn:koi-net.node` providers, reaching out to first contact {self.config.koi_net.first_contact.rid!r}")
            self.handshaker.handshake_with(self.config.koi_net.first_contact.rid)
            
    def stop(self):
        """Stops a node, send stop signals to workers."""
        
        self.kobj_queue.q.put(STOP_WORKER)
        self.event_queue.q.put(STOP_WORKER)


import time
from rid_lib.types import KoiNetNode

from ..protocol.event import Event


class EventBuffer:
    """Stores outgoing events intended for polling nodes."""
    buffers: dict[KoiNetNode, list[Event]]
    start_time: dict[KoiNetNode, float]
    
    def __init__(self):
        self.buffers = {}
        self.start_time = {}
        
    def push(self, node: KoiNetNode, event: Event):
        """Pushes event to specified node.
        
        Sets start time to now if unset.
        """
        
        if node not in self.buffers:
            self.start_time[node] = time.time()
        
        event_buf = self.buffers.setdefault(node, [])
        event_buf.append(event)
        
    def buf_len(self, node: KoiNetNode):
        """Returns the length of a node's event buffer."""
        return len(self.buffers.get(node, []))
        
    def flush(self, node: KoiNetNode, limit: int = 0) -> list[Event]:
        """Flushes all (or limit) events for a node.
        
        Resets start time.
        """
        
        if node in self.start_time:
            del self.start_time[node]
        
        if node not in self.buffers:
            return []
        
        event_buf = self.buffers[node]
        
        if limit and len(event_buf) > limit:
            flushed_events = event_buf[:limit]
            self.buffers[node] = event_buf[limit:]
        else:
            flushed_events = event_buf.copy()
            del self.buffers[node]
        
        return flushed_events


import inspect
from enum import StrEnum
from typing import Any, Protocol, Self
from dataclasses import make_dataclass

import structlog
from pydantic import BaseModel

from .entrypoints.base import EntryPoint

log = structlog.stdlib.get_logger()


class CompType(StrEnum):
    FACTORY = "FACTORY"
    OBJECT = "OBJECT"

class BuildOrderer(type):
    def __new__(cls, name: str, bases: tuple, dct: dict[str]):
        """Sets `cls._build_order` from component order in class definition."""
        cls = super().__new__(cls, name, bases, dct)
        
        if "_build_order" not in dct:
            components: dict[str, Any] = {}
            # adds components from base classes (including cls)
            for base in reversed(inspect.getmro(cls)[:-1]):
                for k, v in vars(base).items():
                    # excludes built in and private attributes
                    if not k.startswith("_"):
                        components[k] = v
                        
            # recipe list constructed from names of non-None components
            cls._build_order = [
                name for name, _type in components.items()
                if _type is not None
            ]
            
        return cls

class NodeContainer(Protocol):
    """Dummy 'shape' for node containers built by assembler."""
    entrypoint = EntryPoint

class NodeAssembler(metaclass=BuildOrderer):
    # Self annotation lying to type checker to reflect typing set in node blueprints
    def __new__(self) -> Self:
        """Returns assembled node container."""
        return self._build_node()
    
    @classmethod
    def _build_deps(cls) -> dict[str, tuple[CompType, list[str]]]:
        """Returns dependency graph for components defined in `cls_build_order`.
        
        Graph representation is a dict where each key is a component name,
        and the value is tuple containing the component type, and a list
        of dependency component names.
        """
        
        dep_graph = {}
        for comp_name in cls._build_order:
            try:
                comp = getattr(cls, comp_name)
            except AttributeError:
                raise Exception(f"Component '{comp_name}' not found in class definition")
            
            if not callable(comp):
                comp_type = CompType.OBJECT
                dep_names = []
            
            elif isinstance(comp, type) and issubclass(comp, BaseModel):
                comp_type = CompType.OBJECT
                dep_names = []
            
            else:
                sig = inspect.signature(comp)
                comp_type = CompType.FACTORY
                dep_names = list(sig.parameters)
                
            dep_graph[comp_name] = (comp_type, dep_names)
            
        return dep_graph
        
    @classmethod
    def _visualize(cls) -> str:
        """Returns representation of dependency graph in Graphviz DOT language."""
        dep_graph = cls._build_deps()
        
        s = "digraph G {\n"
        for node, (_, neighbors) in dep_graph.items():
            sub_s = node
            if neighbors:
                sub_s += f"-> {', '.join(neighbors)}"
            sub_s = sub_s.replace("graph", "graph_") + ";"
            s += " " * 4 + sub_s + "\n"
        s += "}"
        return s
        
    @classmethod
    def _build_comps(cls) -> dict[str, Any]:
        """Returns assembled components from dependency graph."""
        dep_graph = cls._build_deps()
        
        components: dict[str, Any] = {}
        for comp_name, (comp_type, dep_names) in dep_graph.items():
            comp = getattr(cls, comp_name, None)
            
            if comp_type == CompType.OBJECT:
                components[comp_name] = comp
            
            elif comp_type == CompType.FACTORY:
                # builds depedency dict for current component
                dependencies = {}
                for dep_name in dep_names:
                    if dep_name not in components:
                        raise Exception(f"Couldn't find required component '{dep_name}'")
                    dependencies[dep_name] = components[dep_name]
                components[comp_name] = comp(**dependencies)
                
        return components

    @classmethod
    def _build_node(cls) -> NodeContainer:
        """Returns node container from components."""
        components = cls._build_comps()
        
        NodeContainer = make_dataclass(
            cls_name="NodeContainer",
            fields=[
                (name, type(component)) 
                for name, component
                in components.items()
            ],
            frozen=True
        )
        
        return NodeContainer(**components)



"""Defines KOI-net protocol errors."""

from enum import StrEnum


class ErrorType(StrEnum):
    UnknownNode = "unknown_node"
    InvalidKey = "invalid_key"
    InvalidSignature = "invalid_signature"
    InvalidTarget = "invalid_target"

class ProtocolError(Exception):
    error_type: ErrorType
    
class UnknownNodeError(ProtocolError):
    error_type = ErrorType.UnknownNode
    
class InvalidKeyError(ProtocolError):
    error_type = ErrorType.InvalidKey
    
class InvalidSignatureError(ProtocolError):
    error_type = ErrorType.InvalidSignature

class InvalidTargetError(ProtocolError):
    error_type = ErrorType.InvalidTarget


import os
from pydantic import BaseModel, model_validator
from dotenv import load_dotenv
from rid_lib import RIDType
from rid_lib.types import KoiNetNode
import structlog

from koi_net.protocol.secure import PrivateKey
from ..protocol.node import NodeProfile

log = structlog.stdlib.get_logger()


class EventWorkerConfig(BaseModel):
    queue_timeout: float = 0.1
    max_buf_len: int = 5
    max_wait_time: float = 1.0
    
class KobjWorkerConfig(BaseModel):
    queue_timeout: float = 0.1

class NodeContact(BaseModel):
    rid: KoiNetNode | None = None
    url: str | None = None

class KoiNetConfig(BaseModel):
    """Config for KOI-net parameters."""
    
    node_name: str
    node_rid: KoiNetNode | None = None
    node_profile: NodeProfile
    
    rid_types_of_interest: list[RIDType] = [KoiNetNode]
        
    cache_directory_path: str = ".rid_cache"
    private_key_pem_path: str = "priv_key.pem"
    
    event_worker: EventWorkerConfig = EventWorkerConfig()
    kobj_worker: KobjWorkerConfig = KobjWorkerConfig()
    
    first_contact: NodeContact = NodeContact()
    
class EnvConfig(BaseModel):
    """Config for environment variables.
    
    Values set in the config are the variables names, and are loaded
    from the environment at runtime. For example, if the config YAML
    sets `priv_key_password: "PRIV_KEY_PASSWORD"` accessing 
    `priv_key_password` would retrieve the value of `PRIV_KEY_PASSWORD`
    from the environment variables.
    """
    
    priv_key_password: str = "PRIV_KEY_PASSWORD"
    
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        load_dotenv()
    
    def __getattribute__(self, name):
        value = super().__getattribute__(name)
        if name in type(self).model_fields:
            env_val = os.getenv(value)
            if env_val is None:
                raise ValueError(f"Required environment variable {value} not set")
            return env_val
        return value

class NodeConfig(BaseModel):
    """Base node config class, intended to be extended."""
    
    koi_net: KoiNetConfig
    env: EnvConfig = EnvConfig()
    
    @model_validator(mode="after")
    def generate_rid_cascade(self):
        """Generates node RID if missing."""
        if self.koi_net.node_rid and self.koi_net.node_profile.public_key:
            return self
        
        log.debug("Node RID or public key not found in config, attempting to generate")
        
        try:
            # attempts to read existing private key PEM file
            with open(self.koi_net.private_key_pem_path, "r") as f:
                priv_key_pem = f.read()
                priv_key = PrivateKey.from_pem(
                    priv_key_pem,
                    password=self.env.priv_key_password)
                log.debug("Used existing private key from PEM file")
        
        except FileNotFoundError:
            # generates new private key if PEM not found
            priv_key = PrivateKey.generate()
            
            with open(self.koi_net.private_key_pem_path, "w") as f:
                f.write(priv_key.to_pem(self.env.priv_key_password))
            log.debug("Generated new private key, no PEM file found")
        
        pub_key = priv_key.public_key()
        self.koi_net.node_rid = pub_key.to_node_rid(self.koi_net.node_name)
        log.debug(f"Node RID set to {self.koi_net.node_rid}")
        
        if self.koi_net.node_profile.public_key != pub_key.to_der():
            if self.koi_net.node_profile.public_key:
                log.warning("New private key overwriting old public key!")
            
            self.koi_net.node_profile.public_key = pub_key.to_der()
        
        return self


import sys
import logging
from logging.handlers import RotatingFileHandler
from datetime import datetime
import structlog
import colorama


console_renderer = structlog.dev.ConsoleRenderer(
    columns=[
        # Render the timestamp without the key name in yellow.
        structlog.dev.Column(
            "timestamp",
            structlog.dev.KeyValueColumnFormatter(
                key_style=None,
                value_style=colorama.Style.DIM,
                reset_style=colorama.Style.RESET_ALL,
                value_repr=lambda t: datetime.fromisoformat(t).strftime("%Y-%m-%d %H:%M:%S"),
            ),
        ),
        structlog.dev.Column(
            "level",
            structlog.dev.LogLevelColumnFormatter(
                level_styles={
                    level: colorama.Style.BRIGHT + color
                    for level, color in {
                        "critical": colorama.Fore.RED,
                        "exception": colorama.Fore.RED,
                        "error": colorama.Fore.RED,
                        "warn": colorama.Fore.YELLOW,
                        "warning": colorama.Fore.YELLOW,
                        "info": colorama.Fore.GREEN,
                        "debug": colorama.Fore.GREEN,
                        "notset": colorama.Back.RED,
                    }.items()
                },
                reset_style=colorama.Style.RESET_ALL,
                width=9
            )
        ),
        # Render the event without the key name in bright magenta.
        
        # Default formatter for all keys not explicitly mentioned. The key is
        # cyan, the value is green.
        structlog.dev.Column(
            "path",
            structlog.dev.KeyValueColumnFormatter(
                key_style=None,
                value_style=colorama.Fore.MAGENTA,
                reset_style=colorama.Style.RESET_ALL,
                value_repr=str,
                width=30
            ),
        ),
        # structlog.dev.Column(
        #     "func_name",
        #     structlog.dev.KeyValueColumnFormatter(
        #         key_style=None,
        #         value_style=colorama.Fore.MAGENTA,
        #         reset_style=colorama.Style.RESET_ALL,
        #         value_repr=str,
        #         prefix="(",
        #         postfix=")",
        #         width=15
        #     ),
        # ),
        structlog.dev.Column(
            "event",
            structlog.dev.KeyValueColumnFormatter(
                key_style=None,
                value_style=colorama.Fore.WHITE,
                reset_style=colorama.Style.RESET_ALL,
                value_repr=str,
                width=30
            ),
        ),
        structlog.dev.Column(
            "",
            structlog.dev.KeyValueColumnFormatter(
                key_style=colorama.Fore.BLUE,
                value_style=colorama.Fore.GREEN,
                reset_style=colorama.Style.RESET_ALL,
                value_repr=str,
            ),
        )
    ]
)

class LogSystem:
    """Handles initializing the logging system."""
    
    def __init__(self):
        file_handler = RotatingFileHandler(
            filename="log.ndjson",
            maxBytes=10 * 1024 * 1024,
            backupCount=5,
            encoding="utf-8"
        )
        file_handler.setFormatter(
            structlog.stdlib.ProcessorFormatter(
                processor=structlog.processors.JSONRenderer()
            )
        )
        
        console_handler = logging.StreamHandler(sys.stdout)
        console_handler.setFormatter(
            structlog.stdlib.ProcessorFormatter(
                processor=console_renderer
            )
        )
        
        logging.basicConfig(
            level=logging.DEBUG,
            handlers=[file_handler, console_handler]
        )
        
        structlog.configure(
            processors=[
                structlog.stdlib.filter_by_level,
                structlog.stdlib.add_logger_name,
                structlog.stdlib.add_log_level,
                structlog.stdlib.PositionalArgumentsFormatter(),
                structlog.processors.TimeStamper(fmt="iso"),
                # structlog.processors.StackInfoRenderer(),
                structlog.processors.UnicodeDecoder(),
                structlog.processors.CallsiteParameterAdder({
                    structlog.processors.CallsiteParameter.MODULE,
                    structlog.processors.CallsiteParameter.FUNC_NAME
                }),
                # lambda _, __, event: {
                #     **event, 
                #     "path": event["module"] + "." + event["func_name"]
                # },
                # console_renderer
                structlog.stdlib.ProcessorFormatter.wrap_for_formatter
                
            ],
            wrapper_class=structlog.stdlib.BoundLogger,
            logger_factory=structlog.stdlib.LoggerFactory(),
            cache_logger_on_first_use=True,
        )


from koi_net.config.partial_node import PartialNodeConfig, KoiNetConfig, NodeProfile
from koi_net.core import PartialNode


class MyPartialNodeConfig(PartialNodeConfig):
    koi_net: KoiNetConfig = KoiNetConfig(
        node_name="partial",
        node_profile=NodeProfile()
    )

class MyPartialNode(PartialNode):
    config_schema = MyPartialNodeConfig

if __name__ == "__main__":
    node = MyPartialNode()
    node.entrypoint.run()


from ruamel.yaml import YAML

from koi_net.config.proxy import ConfigProxy
from .core import NodeConfig


class ConfigLoader:
    """Loads node config from a YAML file, and proxies access to it."""
    
    file_path: str = "config.yaml"
    file_content: str
    
    config_schema: type[NodeConfig]
    proxy: ConfigProxy
    
    def __init__(
        self, 
        config_schema: type[NodeConfig],
        config: ConfigProxy
    ):
        self.config_schema = config_schema
        self.proxy = config
    
    def load_from_yaml(self):
        """Loads config from YAML file, or generates it if missing."""
        yaml = YAML()
        
        try:
            with open(self.file_path, "r") as f:
                self.file_content = f.read()
            config_data = yaml.load(self.file_content)
            self.proxy._config = self.config_schema.model_validate(config_data)
        
        except FileNotFoundError:
            self.proxy._config = self.config_schema()
        
    def save_to_yaml(self):
        """Saves config to YAML file."""
        yaml = YAML()
        
        with open(self.file_path, "w") as f:
            try:
                config_data = self.proxy._config.model_dump(mode="json")
                yaml.dump(config_data, f)
                
            except Exception as e:
                # rewrites original content if YAML dump fails
                if self.file_content:
                    f.seek(0)
                    f.truncate()
                    f.write(self.file_content)
                raise e